from __future__ import annotations

import importlib.util
import warnings
from typing import Any, Generic, Literal

from numpy.random import RandomState
from sklearn.model_selection import train_test_split
from typing_extensions import Self

from .estimator_wrapper import EstimatorWrapperBase
from .types import TEstimator


class EvalSetWrapper(EstimatorWrapperBase[TEstimator], Generic[TEstimator]):
    """A wrapper that splits the data into train and validation sets and
    passes the validation set to `eval_set` parameter of the estimator."""

    def __init__(
        self,
        estimator: TEstimator,
        *,
        test_size: float | int | None = None,
        train_size: float | int | None = None,
        random_state: int | RandomState | None = None,
        shuffle: bool = True,
        stratify: bool = False,
        **kwargs: Any,
    ) -> None:
        """A wrapper that splits the data into train and validation sets and
        passes the validation set to `eval_set` parameter of the estimator.

        Parameters
        ----------
        estimator : Any
            The estimator to wrap.
        test_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. If None, the value is set to the
            complement of the train size. If ``train_size`` is also None, it will
            be set to 0.25.
            Alias: ``validation_fraction``
        train_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the train split. If
            int, represents the absolute number of train samples. If None,
            the value is automatically set to the complement of the test size.
            Alias: ``train_fraction``
        random_state : int, RandomState instance or None, default=None
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.
        shuffle : bool, optional
            Whether or not to shuffle the data before splitting. If shuffle=False
            then stratify must be None. by default True
        stratify : bool, optional
            Whether or not stratify the data before splitting. If stratify=True,
            y must be categorical. by default False
        **kwargs : Any
            ``validation_fraction`` : alias for ``test_size`` (``sklearn`` style)
            ``train_fraction`` : alias for ``train_size`` (``sklearn`` style)
        """
        super().__init__(estimator)
        self.test_size = test_size
        self.train_size = train_size
        self.random_state = random_state
        self.shuffle = shuffle
        self.stratify = stratify
        self.kwargs = kwargs
        for key, value in kwargs.items():
            if key not in [
                "validation_fraction",
                "train_fraction",
            ]:
                warnings.warn(f"Unknown parameter: {key}: {value}")

    def fit(self, X: Any, y: Any, **fit_params: Any) -> Self:
        """Fit the estimator with `eval_set` set to the validation set.

        Parameters
        ----------
        X : Any
            The training input samples.
        y : Any
            The target values.

        Returns
        -------
        Self
            The fitted estimator.
        """
        X_train, X_test, y_train, y_test = train_test_split(
            X,
            y,
            test_size=self.test_size or self.kwargs.get("validation_fraction", None),
            train_size=self.train_size or self.kwargs.get("train_fraction", None),
            random_state=self.random_state,
            shuffle=self.shuffle,
            stratify=y if self.stratify else None,
        )
        fit_params = fit_params | {
            "eval_set": [(X_test, y_test)],
        }
        self.estimator.fit(X_train, y_train, **fit_params)
        return self


if importlib.util.find_spec("catboost") is not None:
    import re
    from typing import TypeVar

    import tqdm
    from catboost import CatBoost

    TCatBoost = TypeVar("TCatBoost", bound=CatBoost)

    class CatBoostProgressBarWrapper(
        EstimatorWrapperBase[TEstimator], Generic[TEstimator]
    ):
        """A wrapper that splits the data into train and validation sets and
        passes the validation set to `eval_set` parameter of the estimator and
        shows the progress bar using `tqdm`.

        It is recommended to set `iterations` in `CatBoost.__init__` to show the
        progress bar.
        It is recommended to set `early_stopping_rounds` in `CatBoost.__init__`
        to enable early stopping."""

        def __init__(
            self,
            estimator: TEstimator,
            *,
            tqdm_cls: Literal[
                "auto",
                "autonotebook",
                "std",
                "notebook",
                "asyncio",
                "keras",
                "dask",
                "tk",
                "gui",
                "rich",
                "contrib.slack",
                "contrib.discord",
                "contrib.telegram",
                "contrib.bells",
            ]
            | type[tqdm.std.tqdm] = "auto",
            tqdm_kwargs: dict[str, Any] | None = None,
            verbose: bool = True,
        ) -> None:
            """It is recommended to set `iterations` in `CatBoost.__init__` to show the
            progress bar.

            Parameters
            ----------
            estimator : Any
                The estimator to wrap.
            tqdm_cls : Literal['auto', 'autonotebook', 'std', 'notebook', 'asyncio',
                'keras', 'dask', 'tk', 'gui', 'rich', 'contrib.slack', 'contrib.discord',
                'contrib.telegram', 'contrib.bells'] or type[tqdm.std.tqdm] or None, optional
                The tqdm class or module name, by default 'auto'
            tqdm_kwargs : dict[str, Any] or None, optional
                The keyword arguments passed to the tqdm class initializer
            verbose : bool, optional
                Whether to show other logs, by default True

            Examples
            --------
            >>> from catboost import CatBoostRegressor
            >>> from sklearn.datasets import make_regression
            >>> from sklearn_utilities import CatBoostProgressBarWrapper, EvalSetWrapper
            >>> X, y = make_regression()
            >>> # `iterations` is recommended to be set to show the progress bar.
            >>> # `early_stopping_rounds` should be set to enable early stopping.
            >>> estimator = CatBoostRegressor(iterations=1000, early_stopping_rounds=10)
            >>> estimator = EvalSetWrapper(estimator)
            >>> estimator = CatBoostProgressBarWrapper(estimator)
            >>> estimator.fit(X, y)
            """
            super().__init__(
                estimator,
            )
            self.tqdm_cls = tqdm_cls
            self.tqdm_kwargs = tqdm_kwargs
            self.verbose = verbose

            if isinstance(tqdm_cls, str):
                tqdm_module = importlib.import_module(f"tqdm.{tqdm_cls}")
                self.tqdm_cls_ = getattr(tqdm_module, "tqdm")
            else:
                self.tqdm_cls_ = tqdm_cls
            self.tqdm_kwargs_ = tqdm_kwargs or {}
            if "total" in self.tqdm_kwargs_:
                warnings.warn("'total' in tqdm_kwargs is ignored.", UserWarning)

        def fit(self, X: Any, y: Any, **fit_params: Any) -> Self:
            class ProgressBarPrint:
                def __init__(self_child) -> None:
                    self_child.pbar: tqdm.std.tqdm | None = None

                def write(self_child, text: str) -> None:
                    if self_child.pbar is None:
                        params = self.estimator._get_params()
                        iterations = (
                            params.get("iterations", None)
                            or params.get("n_estimators", None)
                            or params.get("num_boost_round", None)
                            or params.get("num_trees", None)
                            or params.get("num_iterations", None)
                        )  # noqa
                        self_child.pbar = self.tqdm_cls_(
                            **(
                                self.tqdm_kwargs_
                                | {
                                    "total": iterations,
                                }
                            ),
                        )
                    try:
                        # 0:      learn: 221.7751345      test: 210.0125818       test1: 210.0125818
                        # best: 210.0125818 (0)   total: 196ms    remaining: 19.4s
                        if not re.match(r"^\d+:", text):
                            raise Exception("Text does not seem to be metrics")
                        text = re.sub(r"[^\s]*: ", "", text)
                        numbers = [
                            float(x) for x in re.findall(r"[-+]?\d*\.\d+|\d+", text)
                        ]
                        n_iter = int(numbers[0]) + 1
                        if "(" in text:
                            # early stopping
                            scores = numbers[1:-4]
                            best_score = numbers[-4]
                            best_iter = int(numbers[-3])
                            self_child.pbar.set_postfix_str(
                                f"{', '.join(f'{score:g}' for score in scores)}"
                                f"{'=' if best_score == scores[1] else '<' if best_score > scores[1] else '>'}"  # noqa
                                f"{best_score:g}@{best_iter:g}it",
                            )
                        else:
                            scores = numbers[1:-2]
                            self_child.pbar.set_postfix_str(
                                f"{', '.join(f'{score:g}' for score in scores)}"
                            )

                        self_child.pbar.update(n_iter - self_child.pbar.n)
                    except Exception:
                        if self.verbose:
                            self_child.pbar.write(text)

            fit_params = fit_params | {
                "log_cout": ProgressBarPrint(),
            }
            self.estimator.fit(X, y, **fit_params)
            return self

    if __name__ == "__main__":
        from catboost import CatBoostRegressor
        from sklearn.datasets import make_regression
        from sklearn.metrics import mean_squared_error

        X, y = make_regression(n_targets=2, random_state=0)
        X_train, X_test, y_train, y_test = train_test_split(X, y)
        estimator = CatBoostProgressBarWrapper(
            EvalSetWrapper(
                CatBoostRegressor(
                    iterations=100,
                    learning_rate=0.4,
                    early_stopping_rounds=10,
                    objective="MultiRMSE",
                ),
            )
        )
        estimator.fit(X_train, y_train)
        mean_squared_error(y_test, estimator.predict(X_test))



from __future__ import annotations

from typing import Any, TypeVar

from numpy.typing import NDArray
from pandas import DataFrame, Series

TEstimator = TypeVar("TEstimator", bound=Any)
TEstimatorVar = TypeVar("TEstimatorVar", bound=Any)
TTransformer = TypeVar("TTransformer", bound=Any)

TX = TypeVar("TX", DataFrame, NDArray[Any])
TY = TypeVar("TY", DataFrame, Series, NDArray[Any])
TXPandas = TypeVar("TXPandas", bound=DataFrame)
TYPandas = TypeVar("TYPandas", DataFrame, Series)



from __future__ import annotations

import numpy as np
from numpy.random import RandomState
from pandas import DataFrame, Series

from .types import TX, TY


def add_missing_values(
    dataset: tuple[TX, TY],
    *,
    missing_rate_x: float = 0.5,
    missing_rate_y: float = 0.5,
    random_state: RandomState | int | None = None,
) -> tuple[TX, TY]:
    """Add missing values to a dataset.

    Parameters
    ----------
    dataset : tuple[TX, TY]
        The dataset to add missing values to.
    missing_rate_x : float, optional
        The rate of missing values to add to X, by default 0.5
    missing_rate_y : float, optional
        The rate of missing values to add to y, by default 0.5
    random_state : RandomState | int | None, optional
        The random state to use, by default None

    Returns
    -------
    tuple[TX, TY]
        The dataset with missing values added.
    """
    random_state = RandomState(random_state)
    X_full_, y_full_ = dataset
    if isinstance(X_full_, DataFrame):
        X_full = X_full_.to_numpy()
    else:
        X_full = X_full_
    if isinstance(y_full_, (DataFrame, Series)):
        y_full = y_full_.to_numpy()
    else:
        y_full = y_full_

    X_mask = random_state.rand(*X_full.shape) < missing_rate_x
    y_mask = random_state.rand(*y_full.shape) < missing_rate_y
    X_missing = X_full.copy()
    X_missing[X_mask] = np.nan
    y_missing = y_full.copy()
    y_missing[y_mask] = np.nan

    if isinstance(X_full_, DataFrame):
        X_missing = DataFrame(X_missing, columns=X_full_.columns, index=X_full_.index)
    if isinstance(y_full_, Series):
        y_missing = Series(y_missing, index=y_full_.index, name=y_full_.name)
    elif isinstance(y_full_, DataFrame):
        y_missing = DataFrame(y_missing, columns=y_full_.columns, index=y_full_.index)
    return X_missing, y_missing



from __future__ import annotations

from typing import Any, Literal

from sklearn.discriminant_analysis import StandardScaler


class StandardScalerVar(StandardScaler):
    with_mean: bool

    def __init__(
        self,
        *,
        copy: bool = True,
        with_mean: bool = True,
        with_std: bool = True,
        var_type: Literal["std", "var"] = "std",
    ) -> None:
        super().__init__(copy=copy, with_mean=with_mean, with_std=with_std)
        self.var_type = var_type

    def inverse_transform(
        self, X: Any, copy: bool | None = None, return_std: bool = False
    ) -> Any:
        if return_std:
            if isinstance(X, tuple):
                return self.inverse_transform(
                    X[0], copy, False
                ), self.inverse_transform(X[1], copy, True)
            prev_with_mean = self.with_mean
            self.with_mean = False
            for i in range(1 if self.var_type == "std" else 2):
                X_scaled = super().inverse_transform(X, copy if i == 0 else None)
            self.with_mean = prev_with_mean
            return X_scaled
        return super().inverse_transform(X, copy)



from __future__ import annotations

from typing import Any, Callable, Sequence

from pandas import DataFrame
from sklearn.base import BaseEstimator, TransformerMixin, check_is_fitted
from typing_extensions import Self

from ..id_transformer import IdTransformer


class ExcludedColumnTransformerPandas(BaseEstimator, TransformerMixin):
    """A transformer that excludes columns from the input data frame."""

    feature_names_in_: Sequence[str]
    feature_names_out_: Sequence[str]

    def __init__(
        self,
        estimator: Any = IdTransformer(),
        exclude_columns: Sequence[str] | Callable[[Sequence[str]], Sequence[bool]] = [],
    ) -> None:
        """A transformer that excludes columns from the input data frame.

        Parameters
        ----------
        estimator : Any, optional
            The estimator to be wrapped, by default IdTransformer()
        exclude_columns : Sequence[str] | Callable[[Sequence[str]], Sequence[bool]],
        optional
            The columns to be excluded, by default []
            If callable, a function that takes the column names as an argument.
        """
        self.exclude_columns = exclude_columns
        self.estimator = estimator

    def _fit(self, X: DataFrame) -> None:
        self.feature_names_in_ = list(X.columns)
        if isinstance(self.exclude_columns, Callable):  # type: ignore
            exclude_columns = X.columns[self.exclude_columns(X.columns)]  # type: ignore
        else:
            exclude_columns = self.exclude_columns
        self.feature_names_out_ = list(set(X.columns) - set(exclude_columns))

    def fit(self, X: DataFrame, **fit_params: Any) -> Self:
        self._fit(X)
        self.estimator.fit(X[self.feature_names_out_], **fit_params)
        return self

    def transform(
        self, X: DataFrame, y: Any = None, **transform_params: Any
    ) -> DataFrame:
        check_is_fitted(self)
        return self.estimator.transform(X[self.feature_names_out_], **transform_params)

    def fit_transform(
        self, X: DataFrame, y: Any = None, **fit_params: Any
    ) -> DataFrame:
        self._fit(X)
        return self.estimator.fit_transform(X[self.feature_names_out_], y, **fit_params)


class IncludedColumnTransformerPandas(BaseEstimator, TransformerMixin):
    """A transformer that includes columns from the input data frame."""

    feature_names_in_: Sequence[str]
    feature_names_out_: Sequence[str]

    def __init__(
        self,
        estimator: Any = IdTransformer(),
        include_columns: Sequence[str] | Callable[[Sequence[str]], Sequence[bool]] = [],
    ) -> None:
        """A transformer that includes columns from the input data frame.

        Parameters
        ----------
        estimator : Any, optional
            The estimator to be wrapped, by default IdTransformer()
        include_columns : Sequence[str] | Callable[[Sequence[str]], Sequence[bool]],
        optional
            The columns to be included, by default []
            If callable, a function that takes the column names as an argument.
        """
        self.include_columns = include_columns
        self.estimator = estimator

    def _fit(self, X: DataFrame) -> None:
        self.feature_names_in_ = list(X.columns)
        if isinstance(self.include_columns, Callable):  # type: ignore
            self.feature_names_out_ = X.columns[
                self.include_columns(X.columns)  # type: ignore
            ]
        else:
            self.feature_names_out_ = self.include_columns  # type: ignore

    def fit(self, X: DataFrame, **fit_params: Any) -> Self:
        self._fit(X)
        self.estimator.fit(X[self.feature_names_out_], **fit_params)
        return self

    def transform(
        self, X: DataFrame, y: Any = None, **transform_params: Any
    ) -> DataFrame:
        check_is_fitted(self)
        return self.estimator.transform(X[self.feature_names_out_], **transform_params)

    def fit_transform(
        self, X: DataFrame, y: Any = None, **fit_params: Any
    ) -> DataFrame:
        self._fit(X)
        return self.estimator.fit_transform(X[self.feature_names_out_], y, **fit_params)



from __future__ import annotations

import warnings
from typing import Any, Callable, Generic

from pandas import Series
from typing_extensions import Self

from ..estimator_wrapper import EstimatorWrapperBase
from ..id_transformer import IdTransformer
from ..types import TX, TY, TEstimator, TTransformer


def _parse_2d(func: Callable[..., TY]) -> Callable[..., TY]:
    def wrapper(X: TX, **kwargs: Any) -> TY:
        ndim = X.ndim
        if ndim == 1:
            if isinstance(X, Series):
                X = X.to_frame()
            else:
                X = X[:, None]
        res = func(X, **kwargs)
        if ndim == 1:
            try:
                return res.squeeze(axis=1)
            except ValueError as e:
                warnings.warn(f"Failed to squeeze: {e}")
        return res

    return wrapper


class TransformedTargetEstimatorVar(
    EstimatorWrapperBase[TEstimator], Generic[TEstimator, TTransformer]
):
    """TransformTargetRegressor with std/var support."""

    def __init__(
        self,
        estimator: TEstimator,
        *,
        transformer: TTransformer = IdTransformer(),
        inverse_transform_separately: bool = False,
    ) -> None:
        super().__init__(estimator)
        self.transformer = transformer
        self.inverse_transform_separately = inverse_transform_separately

    def fit(self, X: TX, y: TY, **fit_params: Any) -> Self:
        y = _parse_2d(self.transformer.fit_transform)(y)
        self.estimator.fit(X, y, **fit_params)
        return self

    def predict(self, X: TX, **predict_params: Any) -> TY | tuple[TY, TY]:
        if predict_params.get("return_std", False):
            pred, pred_std = self.estimator.predict(X, **predict_params)
            if self.inverse_transform_separately:
                return _parse_2d(self.transformer.inverse_transform)(pred), _parse_2d(
                    self.transformer.inverse_transform
                )(pred_std, return_std=True)
            else:
                return _parse_2d(self.transformer.inverse_transform)(
                    (pred, pred_std), return_std=True
                )
        pred = self.estimator.predict(X, **predict_params)
        return _parse_2d(self.transformer.inverse_transform)(pred)

    def predict_var(self, X: TX, **predict_params: Any) -> TY:
        pred_var = self.estimator.predict_var(X, **predict_params)
        return _parse_2d(self.transformer.inverse_transform_var)(pred_var)



from __future__ import annotations

from typing import Any, Generic, Hashable, Sequence

from pandas import DataFrame, Series, concat
from sklearn.base import BaseEstimator, TransformerMixin, check_is_fitted
from typing_extensions import Self

from .pandas import DataFrameWrapper
from .types import TEstimator


class AppendXPredictionToX(BaseEstimator, TransformerMixin, Generic[TEstimator]):
    """Append the prediction of X by the estimator to X."""

    estimator: TEstimator

    def __init__(
        self,
        estimator: TEstimator,
        *,
        variables: Sequence[Hashable] | None = None,
        append: bool = True,
        append_pred_diff: bool = True,
        append_pred_real_diff: bool = True,
    ) -> None:
        """Append the prediction of X by the estimator to X.
        The new columns are suffixed with "_pred".

        Parameters
        ----------
        estimator : TEstimator
            The estimator to be wrapped.
        variables : Sequence[Hashable] | None, optional
            The variables to be used for prediction, by default None
        append : bool, optional
            Whether to append the original X, by default True
        append_diff : bool, optional
            Whether to append the difference between
            the previous prediction and the current prediction, by default True
            The column names are suffixed with "_pred_diff"
        append_real_diff : bool, optional
            Whether to append the difference between
            the current value and the current prediction, by default True
            The column names are suffixed with "_pred_real_diff"
        """
        self.estimator = estimator
        self.variables = variables
        self.append = append
        self.append_pred_diff = append_pred_diff
        self.append_pred_real_diff = append_pred_real_diff

    def fit(self, X: DataFrame, y: Series | None = None, **fit_params: Any) -> Self:
        if self.variables is not None:
            X = X.loc[:, self.variables]
        X_future = X.shift(-1)  # do not specify freq

        # drop first
        X = X.iloc[:-1]
        X_future = X_future.iloc[:-1]

        self.estimator_ = DataFrameWrapper(self.estimator)
        self.estimator_.fit(X, X_future, **fit_params)
        return self

    def transform(
        self, X: DataFrame, y: Series | None = None, **transform_params: Any
    ) -> DataFrame:
        check_is_fitted(self, "estimator_")

        to_concat = []

        # X
        if self.append:
            to_concat.append(X)

        # X_pred
        X_variables = X.loc[:, self.variables] if self.variables is not None else X
        X_pred = self.estimator_.predict(X_variables)
        to_concat.append(X_pred.add_suffix("_pred"))

        # X_pred_diff
        if self.append_pred_diff:
            X_pred_diff = X_pred.diff(1)
            to_concat.append(X_pred_diff.add_suffix("_pred_diff"))

        # X_pred_real_diff
        if self.append_pred_real_diff:
            X_pred_real_diff = X_pred - X_variables
            to_concat.append(X_pred_real_diff.add_suffix("_pred_real_diff"))

        # concat
        X_concat = concat(to_concat, axis=1)
        return X_concat



from __future__ import annotations

from typing import Any, Generic

from pandas import DataFrame
from typing_extensions import Self

from .estimator_wrapper import EstimatorWrapperBase
from .types import TEstimator


class DropMissingRowsY(EstimatorWrapperBase[TEstimator], Generic[TEstimator]):
    """A wrapper for estimators that drops NaN values from y before fitting."""

    def __init__(self, estimator: TEstimator) -> None:
        """A wrapper for estimators that drops NaN values from y before fitting.

        Parameters
        ----------
        estimator : TEstimator
            The estimator to be wrapped.
        """
        super().__init__(estimator)

    def fit(self, X: DataFrame, y: Any = None, **fit_params: Any) -> Self:
        y = y.dropna()
        idx = X.index.intersection(y.index)
        self.estimator.fit(X.loc[idx], y.loc[idx], **fit_params)
        return self



from __future__ import annotations

from logging import getLogger
from typing import Any, Generic, TypeVar

import joblib
import numpy as np
from numpy.typing import NDArray
from pandas import DataFrame, Series
from sklearn.base import BaseEstimator, RegressorMixin, clone
from sklearn.utils.validation import check_is_fitted
from tqdm_joblib import tqdm_joblib
from typing_extensions import Self

from ..types import TEstimator
from ..utils import drop_X_y
from .dataframe_wrapper import to_frame_or_series_tuple

LOG = getLogger(__name__)

# Scikit learn compatible esitimator that wraps an estimator
# Creates a new estimator for each target variable
# X and Y are pandas dataframes
# X.index.intersection(y.index) is used to fit the estimator,
# calculate for each column
# not X.intersection(Y.index)
# use joblib Parallel to fit the estimators in parallel

TX = TypeVar("TX", bound="DataFrame | Series")
TY = TypeVar("TY", bound="DataFrame | Series")


def _fit_X_y(
    estimator: Any,
    X: DataFrame,
    y: Series,
    *,
    pass_numpy: bool = False,
    safe: bool = False,
    **fit_params: Any,
) -> BaseEstimator:
    X, y = drop_X_y(X, y)
    LOG.debug(f"Length of {y.name}: {len(y)}")
    estimator = clone(estimator, safe=safe)
    if pass_numpy:
        return estimator.fit(X.values, y.values, **fit_params)
    return estimator.fit(X, y, **fit_params)


class SmartMultioutputEstimator(BaseEstimator, RegressorMixin, Generic[TEstimator]):
    estimator: TEstimator
    estimators_: list[TEstimator]

    def __init__(
        self,
        estimator: TEstimator,
        *,
        n_jobs: int | None = -1,
        verbose: int = 1,
        pass_numpy: bool = False,
    ) -> None:
        """MultioutputEstimator that
        1. Supports tuples of arrays in `predict()` (for
        `return_std=True`)
        2. Returns list of scores in `score()`
        3. Supports pandas DataFrame and Series

        Parameters
        ----------
        estimator : TEstimator
            The estimator to be wrapped.
        n_jobs : int | None, optional
            The number of jobs to run in parallel, by default -1
        verbose : int, optional
            Whether to show progress bar, by default 1
        pass_numpy : bool, optional
            Whether to pass numpy arrays to the estimator, by default False
        """
        self.estimator = estimator
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.pass_numpy = pass_numpy

    def fit(self, X: DataFrame, y: DataFrame, **fit_params: Any) -> Self:
        self.feature_names_in_ = X.columns
        self.y_names_in_ = y.columns
        with tqdm_joblib(
            desc="Fitting estimators",
            total=len(y.columns),
            disable=self.verbose == 0,
        ):
            estimators = joblib.Parallel(n_jobs=self.n_jobs)(
                joblib.delayed(_fit_X_y)(
                    self.estimator,
                    X,
                    item,
                    pass_numpy=self.pass_numpy,
                    **fit_params,
                )
                for col, item in y.items()
            )
            if estimators is None:
                raise ValueError("No estimators fitted")
            self.estimators_ = list(estimators)
        return self

    def predict(
        self, X: DataFrame, **predict_params: Any
    ) -> (
        DataFrame
        | Series
        | NDArray[Any]
        | tuple[DataFrame | Series | NDArray[Any], ...]
    ):
        check_is_fitted(self)
        X = X[self.feature_names_in_]
        preds = [est.predict(X, **predict_params) for est in self.estimators_]
        preds_: DataFrame | Series | NDArray[Any] | tuple[
            DataFrame | Series | NDArray[Any], ...
        ]
        if any(isinstance(pred, tuple) for pred in preds):
            # list of tuples of arrays to tuples of arrays
            preds_ = tuple(np.array(pred).T for pred in zip(*preds))
        else:
            preds_ = np.array(preds).T
        return to_frame_or_series_tuple(preds_, X.index, self.y_names_in_)

    def score(self, X: DataFrame, y: DataFrame, **score_params: Any) -> NDArray[Any]:
        check_is_fitted(self)
        X = X[self.feature_names_in_]
        y = y[self.y_names_in_]
        return np.array(
            [
                est.score(X, y[col], **score_params)
                for est, col in zip(self.estimators_, self.y_names_in_)
            ]
        )

    def predict_var(
        self, X: DataFrame, **predict_params: Any
    ) -> (
        DataFrame
        | Series
        | NDArray[Any]
        | tuple[DataFrame | Series | NDArray[Any], ...]
    ):
        check_is_fitted(self)
        X = X[self.feature_names_in_]
        preds = np.array(
            [est.predict_var(X, **predict_params) for est in self.estimators_]
        ).T
        return to_frame_or_series_tuple(preds, X.index, self.y_names_in_)

    def __iter__(self) -> Any:
        return iter(self.estimators_)

    def __len__(self) -> int:
        return len(self.estimators_)



from __future__ import annotations

import warnings
from typing import Any, Generic, TypeVar

import numpy as np
import torch
from numpy.lib.stride_tricks import sliding_window_view
from skorch import NeuralNet
from typing_extensions import Self

from sklearn_utilities.estimator_wrapper import EstimatorWrapperBase
from sklearn_utilities.types import TX, TY

TEstimator = TypeVar("TEstimator", bound=NeuralNet)


class SkorchReshaper(EstimatorWrapperBase[TEstimator], Generic[TEstimator]):
    """skorch wrapper that reshapes tabular data for NNs."""

    def __init__(self, estimator: TEstimator) -> None:
        """skorch wrapper that reshapes tabular data for NNs.
        X: [B, F] -> [B, F]
        y: [B] -> [B, 1] or [B, NY] -> [B, NY] where
        B: batch, F: features, NY: number of outputs

        Parameters
        ----------
        estimator : TEstimator
            The estimator to wrap.
        """
        super().__init__(estimator)
        if self.estimator.device == "cpu":
            warnings.warn(
                "You are using a CPU for training, which may be very slow. "
                f"Consider using a GPU. CUDA Availability: {torch.cuda.is_available()}",
                UserWarning,
            )

    def fit(self, X: TX, y: TY, **fit_params: Any) -> Self:
        # allow multioutput
        X_, y_ = self._validate_data(
            X,
            y,
            validate_separately=(
                {"force_all_finite": False, "allow_nd": True, "ensure_2d": False},
                {"force_all_finite": False, "allow_nd": True, "ensure_2d": False},
            ),
        )
        X_ = X_.astype(np.float32)
        y_ = y_.astype(np.float32)
        if y_.ndim == 1:
            y_ = np.expand_dims(y_, axis=1)
        self.y_ndim_ = y_.ndim
        self.estimator.fit(X_, y_, **fit_params)
        return self

    def predict(self, X: TX, **predict_params: Any) -> TY:
        X_: np.ndarray = self._validate_data(
            X, force_all_finite=False, allow_nd=True, ensure_2d=False
        )
        X_ = X_.astype(np.float32)
        y = self.estimator.predict(X_, **predict_params)
        if self.y_ndim_ == 1 and y.shape[1] == 1:
            if y.shape[1] == 1:
                y = y.squeeze(axis=1)
        return y


class SkorchCNNReshaper(EstimatorWrapperBase[TEstimator], Generic[TEstimator]):
    """skorch wrapper that reshapes tabular data for CNNs using sliding windows."""

    def __init__(self, estimator: TEstimator, *, window_size: int | None) -> None:
        """skorch wrapper that reshapes tabular data for CNNs using sliding windows.
        X: [B, F] -> [B - H + 1, 1, H, F] if window_size is not None (for Conv2d)
           [B, F] -> [B, 1, F] if window_size is None (for Conv1d)
        y: [B] -> [B - H + 1, 1] or [B, NY] -> [B - H + 1, NY] where
        C = 1: channels, B: batch, H: window, F: features, NY: number of outputs

        Parameters
        ----------
        estimator : TEstimator
            The estimator to wrap.
        window_size : int | None
            The size of the sliding window.
            Make sure that CNN kernel size is equal or larger than this.
            If None, no sliding window is applied.
        """
        self.estimator = estimator
        self.window_size = window_size
        if self.estimator.device == "cpu":
            warnings.warn(
                "You are using a CPU for training, which may be very slow. "
                f"Consider using a GPU. CUDA availability: {torch.cuda.is_available()}",
                UserWarning,
            )

    def fit(self, X: TX, y: TY, **fit_params: Any) -> Self:
        X_, y_ = self._validate_data(
            X,
            y,
            validate_separately=(
                {"force_all_finite": False, "allow_nd": True, "ensure_2d": False},
                {"force_all_finite": False, "allow_nd": True, "ensure_2d": False},
            ),
        )
        X_ = X_.astype(np.float32)
        y_ = y_.astype(np.float32)
        if self.window_size is not None:
            X_ = sliding_window_view(X_, self.window_size, axis=0)
            if X_.shape[0] != y_.shape[0] - self.window_size + 1:
                raise AssertionError(
                    f"X.shape[0] = {X_.shape[0]} "
                    "!= y.shape[0] - self.window_size + 1 = "
                    f"{y_.shape[0] - self.window_size + 1}"
                )
            y_ = y_[self.window_size - 1 :]
        X_ = np.expand_dims(X_, axis=1)
        self.y_ndim_ = y_.ndim
        if y_.ndim == 1:
            y_ = np.expand_dims(y_, axis=1)
        self.estimator.fit(X_, y_, **fit_params)
        return self

    def predict(self, X: TX, **predict_params: Any) -> TY:
        X_ = self._validate_data(
            X, force_all_finite=False, allow_nd=True, ensure_2d=False
        )
        X_ = X_.astype(np.float32)
        if self.window_size is not None:
            X_ = sliding_window_view(X_, self.window_size, axis=0)
        X_ = np.expand_dims(X_, axis=1)
        y = self.estimator.predict(X_, **predict_params)
        if self.y_ndim_ == 1:
            if y.shape[1] == 1:
                y = y.squeeze(axis=1)
        return y



from .types import TXPandas, TYPandas


def intersect_X_y(X: TXPandas, y: TYPandas) -> tuple[TXPandas, TYPandas]:
    idx = X.index.intersection(y.index)
    return X.loc[idx], y.loc[idx]


def drop_X_y(X: TXPandas, y: TYPandas) -> tuple[TXPandas, TYPandas]:
    return intersect_X_y(X, y.dropna())



from functools import partial
from typing import Any
from unittest.mock import patch

from pandas import concat
from sklearn.pipeline import FeatureUnion

_horizontal_concat = partial(concat, axis=1)


class FeatureUnionPandas(FeatureUnion):
    def fit_transform(self, X: Any, y: Any = None, **fit_params: Any) -> Any:
        with patch("numpy.hstack", side_effect=_horizontal_concat):
            return super().fit_transform(X, y, **fit_params)

    def transform(self, X: Any) -> Any:
        with patch("numpy.hstack", side_effect=_horizontal_concat):
            return super().transform(X)



from __future__ import annotations

from typing import Any

import numpy as np
from sklearn.feature_selection import GenericUnivariateSelect
from typing_extensions import Self


class LooseGenericUnivariateSelect(GenericUnivariateSelect):
    """A GenericUnivariateSelect that does not require y
    and accepts missing values in X."""

    def fit(self, X: Any, y: Any = None) -> Self:
        y = np.ones(X.shape[0])
        return super().fit(X, y)

    def _validate_data(self, X: Any, y: Any = None, *args: Any, **kwargs: Any) -> Any:
        kwargs["force_all_finite"] = False
        kwargs.pop("multi_output", None)
        return super()._validate_data(X, y, *args, **kwargs)

    def _more_tags(self) -> dict[str, Any]:
        return {"requires_y": False, "allow_nan": True}


def DropMissingColumns(
    threshold_not_missing: float = 0.9,
) -> LooseGenericUnivariateSelect:
    """Drop columns with missing values above a threshold.

    Parameters
    ----------
    threshold_not_missing : float, optional
        If the ratio of non-missing values is below or equals to this threshold,
        the column is dropped, by default 0.9
    """
    if threshold_not_missing > 1:
        return LooseGenericUnivariateSelect(
            lambda X, y: np.isfinite(X).mean(axis=0),
            mode="k_best",
            param=threshold_not_missing,
        )
    return LooseGenericUnivariateSelect(
        lambda X, y: np.isfinite(X).mean(axis=0),
        mode="percentile",
        param=threshold_not_missing * 100,
    )


# class DropMissingColumns(BaseEstimator, SelectorMixin):
#     """Drop columns with missing values above a threshold."""
#     def __init__(self, threshold_not_missing: float = 0.5) -> None:
#         """Drop columns with missing values above a threshold.

#         Parameters
#         ----------
#         threshold_not_missing : float, optional
#             If the ratio of non-missing values is below this threshold,
#             the column is dropped, by default 0.5
#         """
#         self.threshold_not_missing = threshold_not_missing

#     def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Any:
#         self.selector = ColumnTransformer(
#             [
#                 (
#                     "drop_missing_columns",
#                     "passthrough",
#                     lambda X: X.notna().mean(axis=0) >= self.threshold_not_missing,
#                 )
#             ]
#         )
#         self.selector.fit(X, y, **fit_params)
#         return self

#     def transform(self, X: Any, **transform_params: Any) -> Any:
#         return self.selector.transform(X, **transform_params)



from __future__ import annotations

import inspect
import re
import warnings
from typing import Any, Generic, Hashable, TypeVar

from numpy.typing import NDArray
from pandas import DataFrame, Index, Series

from ..estimator_wrapper import EstimatorWrapperBase, TEstimator

TArray = TypeVar("TArray", bound="DataFrame | Series | NDArray[Any]")


def to_frame_or_series(
    array: TArray,
    base_index: Index[Any],
    base_columns_or_name: Index[Any] | Hashable | None,
) -> DataFrame | Series | TArray:
    if isinstance(array, (DataFrame, Series)):
        return array
    try:
        if array.ndim == 1:
            return Series(
                array,
                index=base_index if array.shape[0] == len(base_index) else None,
                name=base_columns_or_name
                if not isinstance(base_columns_or_name, Index)
                else None,
            )
        if array.ndim == 2:
            return DataFrame(
                array,
                index=base_index if array.shape[0] == len(base_index) else None,
                columns=base_columns_or_name
                if (
                    isinstance(base_columns_or_name, Index)
                    and array.shape[1] == len(base_columns_or_name)
                )
                else None,
            )
    except Exception as e:
        warnings.warn(f"Could not convert {array} to DataFrame or Series: {e}")
        return array
    return array


def to_frame_or_series_tuple(
    array: tuple[TArray, ...] | TArray,
    base_index: Index[Any],
    base_columns_or_name: Index[Any] | Hashable,
) -> tuple[DataFrame | Series | TArray, ...] | DataFrame | Series | TArray:
    if isinstance(array, tuple):
        return tuple(
            to_frame_or_series(a, base_index, base_columns_or_name) for a in array
        )
    return to_frame_or_series(array, base_index, base_columns_or_name)


class DataFrameWrapper(EstimatorWrapperBase[TEstimator], Generic[TEstimator]):
    pattern_x: str
    y_columns_or_name: Index[Any] | Hashable | None = None

    def __init__(
        self,
        estimator: TEstimator,
        *,
        pattern_x: str = "^(:?fit|transform|fit_transform)$",
        pattern_y: str = "^predict.*?$",
    ) -> None:
        """A wrapper for estimators that returns pandas DataFrame or Series
        instead of numpy arrays for the methods that have "X" as an argument
        and the name matches the given pattern.

        Parameters
        ----------
        estimator : Any
            The estimator to be wrapped.
        pattern_x : str, optional
            The regex pattern to match the method names,
            by default "^(:?transform|fit_transform)$"
        pattern_y : str, optional
            The regex pattern to match the method names,
            by default "^predict.*$"
        """
        super().__init__(estimator)
        self.pattern_x = pattern_x
        self.pattern_y = pattern_y

    def _save_y_columns_or_name(self, y: Any) -> None:
        if isinstance(y, Series):
            self.y_columns_or_name = y.name
        elif isinstance(y, DataFrame):
            self.y_columns_or_name = y.columns

    def __getattribute__(self, __name: str) -> Any:
        try:
            # do not call super().__getattribute__
            return object.__getattribute__(self, __name)
        except AttributeError:
            attr = getattr(self.estimator, __name)

            x_match = re.search(self.pattern_x, __name)
            y_match = re.search(self.pattern_y, __name)
            if (
                callable(attr)
                and (x_match or y_match)
                and "X" in inspect.signature(attr).parameters
            ):

                def wrapper(*args: Any, **kwargs: Any) -> Any:
                    # get X to get index and columns
                    X = inspect.signature(attr).bind(*args, **kwargs).arguments["X"]

                    # save y columns or name
                    if "y" in inspect.signature(attr).parameters:
                        y = inspect.signature(attr).bind(*args, **kwargs).arguments["y"]
                        self._save_y_columns_or_name(y)

                    # get result
                    result = attr(*args, **kwargs)

                    # avoid fit() not returning self but self.estimator
                    if result is self.estimator:
                        return self

                    # support tuple for return_std=True, etc.
                    return to_frame_or_series_tuple(
                        result,
                        X.index,
                        X.columns if x_match else self.y_columns_or_name,
                    )

                return wrapper

            # behaves like EstimatorWrapperBase
            return attr



from __future__ import annotations

import math
from typing import Any, Generic, Literal, Sequence

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ...types import TX, TY, TEstimator
from .reshaper import SkorchCNNReshaper, SkorchReshaper


class AllowNan(nn.Module):
    """Replaces NaNs in the target values with the predictions."""

    def __init__(
        self, loss: nn.Module, *, has_nan: Literal["left", "right", "both"] = "both"
    ) -> None:
        """Replaces NaNs in the target values with the predictions.

        Parameters
        ----------
        loss : nn.Module
            The loss function to use.
        has_nan : Literal['left', 'right', 'both'], optional
            If 'left', allows NaNs in the first argument of the loss function.
            If 'right', allows NaNs in the second argument of the loss function.
            If 'both', allows NaNs in both arguments of the loss function.
        """
        super().__init__()
        self.loss = loss
        self.has_nan = has_nan

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        if self.has_nan == "both":
            is_nan = torch.isnan(y_pred) | torch.isnan(y_true)
        elif self.has_nan == "left":
            is_nan = torch.isnan(y_pred)
        elif self.has_nan == "right":
            is_nan = torch.isnan(y_true)
        else:
            raise ValueError(f"Unknown has_nan: {self.has_nan}")

        y_pred = torch.where(is_nan, torch.zeros_like(y_pred), y_pred)
        y_true = torch.where(is_nan, torch.zeros_like(y_true), y_true)
        return self.loss(y_pred, y_true)


class LNErrors(nn.Module):
    """Returns L^n errors (not mean of L^n errors)."""

    def __init__(self, n: int = 2) -> None:
        """Returns L^n errors (not mean of L^n errors).

        Parameters
        ----------
        n : int, optional
            The exponent, by default 2
        """
        super().__init__()
        self.n = n

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        return torch.abs(y_true - y_pred).pow(self.n)


class LogCoshErrors(nn.Module):
    """Log cosh errors.
    Loss = log(cosh(errors + eps))

    See also
    --------
    https://datascience.stackexchange.com/questions/96271/logcoshloss-on-pytorch
    """

    def __init__(self, *, softplus: bool = True, eps: float | None = None) -> None:
        """Returns log(cosh(errors)).

        Parameters
        ----------
        softplus : bool, optional
            If True, uses softplus to get stable results, by default True"""
        super().__init__()
        self.softplus = softplus
        self.eps = eps

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        x = y_pred - y_true
        if self.softplus:
            return x + F.softplus(-2.0 * x) - math.log(2.0)
        else:
            eps = self.eps or torch.finfo(x.dtype).eps
            return torch.log(torch.cosh(x + eps))


class XTanhErrors(nn.Module):
    """XTanh errors.
    Loss = x * tanh(x)

    See also
    --------
    https://github.com/tuantle/regression-losses-pytorch
    """

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        x = y_pred - y_true
        return x * torch.tanh(x)


class XSigmoidErrors(nn.Module):
    """XSigmoid errors.
    Loss = x * (2 * sigmoid(x) - 1)

    See also
    --------
    https://github.com/tuantle/regression-losses-pytorch
    """

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        x = y_pred - y_true
        return x * (2 * torch.sigmoid(x) - 1)


class AlgebraicErrors(nn.Module):
    """Algebraic errors.
    Loss = x^2 / sqrt(1 + x^2)

    See also
    --------
    https://github.com/tuantle/regression-losses-pytorch
    """

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        x = y_pred - y_true
        x2 = torch.pow(x, 2)
        return x2 / torch.sqrt(1 + x2)


class AsymmetricLoss(nn.Module):
    def __init__(
        self,
        *,
        t: float,
        loss_pred_grater: nn.Module = LNErrors(1),
        loss_pred_less: nn.Module = LNErrors(1),
    ) -> None:
        super().__init__()
        self.loss_pred_grater = loss_pred_grater
        self.loss_pred_less = loss_pred_less
        self.t = t

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        return torch.mean(
            torch.max(
                self.t * self.loss_pred_grater(y_pred, y_true),
                (1 - self.t) * self.loss_pred_less(y_pred, y_true),
            )
        )


def _reshape_y_pred(
    y_pred: torch.Tensor, y_true_shape: torch.Size, n_ts: int
) -> torch.Tensor:
    if y_pred.ndim == 2 and len(y_true_shape) == 2:
        # [B, Ts * Ny] -> [B, Ts, Ny]
        if y_pred.shape[1] == n_ts * y_true_shape[1]:
            y_pred = y_pred.reshape(y_pred.shape[0], n_ts, -1)
        else:
            raise ValueError(
                "Got y_pred.ndim == 2 and y_true.ndim == 2, but "
                f"y_pred.shape[1] = {y_pred.shape[1]} != "
                f"n_ts * y_true.shape[1] = {n_ts} * {y_true_shape[1]}"
            )
    return y_pred


def _get_ts_axis(y_pred_shape: torch.Size, y_true_shape: torch.Size) -> int:
    # ignore first axis (batch)
    y_pred_shape = y_pred_shape[1:]
    y_true_shape = y_true_shape[1:]

    # find the shape difference between y_pred and y_true
    # if Ts = Ny (y_true_shape[1] = y_true_shape[2]),
    # then ts_axis = -1
    ts_axis = -1
    for i, s in enumerate(y_pred_shape):
        if s not in y_true_shape:
            ts_axis = i + 1
            break
    return ts_axis


class AsymmetricLosses(nn.Module):
    """Asymmetric loss with multioutput support.

    Parameters
    ----------
    y_pred : torch.Tensor
        The predicted values.
        [B, Ts * Ny] or [B, Ts, Ny] or [B, Ny, Ts]
    y_true : torch.Tensor
        The true values.
        [B, Ny] or [B] (if Ny == 1)

    Returns
    -------
    torch.Tensor
        0-dim tensor with the loss."""

    def __init__(
        self, *, ts: Sequence[float] | int, loss: nn.Module = LNErrors(1)
    ) -> None:
        """Asymmetric loss with multioutput support.

        Parameters
        ----------
        ts : Sequence[float] | int
            The list of `t` to use for fitting the data or the number of `t` to use.
            If `ts` is an integer,
            `np.linspace(1 / (ts * 2), 1 - 1 / (ts * 2), ts)` is used.
        loss : nn.Module, optional
            The loss function to use, by default LNErrors(1)
        """
        super().__init__()
        if isinstance(ts, int):
            ts = list(np.linspace(1 / (ts * 2), 1 - 1 / (ts * 2), ts))
        self.ts = ts
        self.loss = loss
        self.losses = nn.ModuleList(
            [
                AsymmetricLoss(t=t, loss_pred_grater=loss, loss_pred_less=loss)
                for t in ts
            ]
        )

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        y_pred = _reshape_y_pred(y_pred, y_true.shape, len(self.ts))
        self.y_true_shape_ = y_true.shape
        self.ts_axis_ = _get_ts_axis(y_pred.shape, y_true.shape)

        # calculate losses
        losses = []
        for y_pred_q, loss in zip(
            torch.moveaxis(y_pred, self.ts_axis_, 0), self.losses
        ):
            losses.append(loss(y_pred_q, y_true))
        return torch.stack(losses).mean()


class SkorchReshaperProbaMixin(Generic[TEstimator]):
    """skorch wrapper mixin that converts quantile predictions to mean and std."""

    estimator: TEstimator
    m_type: Literal["mean", "median", "nanmean", "nanmedian"]
    """M-statistics type to return from `predict` by default"""
    var_type: Literal["var", "std", "ptp", "nanvar", "nanstd"]
    """Variance type to return from `predict` by default"""

    def __init__(
        self,
        estimator: TEstimator,
        *args: Any,
        m_type: Literal["mean", "median", "nanmean", "nanmedian"] = "mean",
        var_type: Literal["var", "std", "ptp", "nanvar", "nanstd"] = "std",
        **kwargs: Any,
    ) -> None:
        super().__init__(estimator, *args, **kwargs)  # type: ignore
        self.m_type = m_type
        self.var_type = var_type

    @staticmethod
    def _get(
        y: TY,
        type_: Literal[
            "mean",
            "median",
            "var",
            "std",
            "ptp",
            "nanmean",
            "nanmedian",
            "nanvar",
            "nanstd",
        ],
        *,
        axis: int,
    ) -> TY:
        if hasattr(np, type_):
            return getattr(np, type_)(y, axis=axis)
        else:
            raise ValueError(f"Unknown type: {type_}")

    def predict(
        self,
        X: TX,
        *,
        return_std: bool = False,
        type_: Literal[
            "mean",
            "median",
            "nanmean",
            "nanmedian",
            "var",
            "std",
            "ptp",
            "nanvar",
            "nanstd",
        ]
        | tuple[
            Literal["mean", "median", "nanmean", "nanmedian"],
            Literal["var", "std", "ptp", "nanvar", "nanstd"],
        ]
        | None = None,
        **predict_params: Any,
    ) -> TY | tuple[TY, TY]:
        ts_axis_ = self.estimator.criterion.ts_axis_
        y = super().predict(X, **predict_params)  # type: ignore
        y = _reshape_y_pred(
            y, self.estimator.criterion.y_true_shape_, len(self.estimator.criterion.ts)
        )

        if return_std or isinstance(type_, tuple):
            if isinstance(type_, str):
                type_tuple = (type_, self.var_type)
            elif type_ is None:
                type_tuple = (self.m_type, self.var_type)
            else:
                type_tuple = type_
            return self._get(y, type_tuple[0], axis=ts_axis_), self._get(
                y, type_tuple[1], axis=ts_axis_
            )
        return self._get(y, type_ or self.m_type, axis=ts_axis_)


class SkorchReshaperProba(  # type: ignore
    SkorchReshaperProbaMixin[TEstimator],
    SkorchReshaper[TEstimator],
    Generic[TEstimator],
):
    def __init__(
        self,
        estimator: TEstimator,
        m_type: Literal["mean", "median", "nanmean", "nanmedian"] = "mean",
        var_type: Literal["var", "std", "ptp", "nanvar", "nanstd"] = "std",
    ) -> None:
        """skorch wrapper that reshapes tabular data for NNs
        and supports quantile predictions.
        X: [B, F] -> [B, F]
        y: [B] -> [B, 1] or [B, NY] -> [B, NY] where
        B: batch, F: features, NY: number of outputs

        Use `AsymmetricLosses` for NeuralNet.criterion.

        `estimator.module` may return [B, Ts * NY] or [B, Ts, NY] or [B, NY, Ts].
        If NY = Ts, assumes that [B, NY, Ts] is returned.

        Parameters
        ----------
        estimator : TEstimator
            The estimator to wrap.
        m_type : Literal['mean', 'median', 'nanmean', 'nanmedian'], optional
            M-statistics type to return from `predict` by default, by default "mean"
        var_type : Literal['var', 'std', 'ptp', 'nanvar', 'nanstd'], optional
            Variance type to return from `predict` by default, by default "std"

        Examples
        --------
        >>> from sklearn_utilities.torch.skorch import SkorchReshaperProba, AsymmetricLosses
        >>> from skorch import NeuralNet
        >>> import torch.nn as nn
        >>> net = nn.Sequential(nn.LazyLinear(10), nn.GELU(), nn.LazyLinear(5))
        >>> est = SkorchReshaperProba(NeuralNet(module=net, criterion=AsymmetricLosses(ts=5)))
        >>> est.fit(X_train, Y_train)
        >>> y_pred, y_std = est.predict(X_test, return_std=True)
        """
        super().__init__(estimator, m_type=m_type, var_type=var_type)


class SkorchCNNReshaperProba(  # type: ignore
    SkorchReshaperProbaMixin[TEstimator],
    SkorchCNNReshaper[TEstimator],
    Generic[TEstimator],
):
    def __init__(
        self,
        estimator: TEstimator,
        *,
        window_size: int | None,
        m_type: Literal["mean", "median", "nanmean", "nanmedian"] = "mean",
        var_type: Literal["var", "std", "ptp", "nanvar", "nanstd"] = "std",
    ) -> None:
        """skorch wrapper that reshapes tabular data for CNNs using sliding windows
        and supports quantile predictions.
        X: [B, F] -> [B - H + 1, 1, H, F] if window_size is not None (for Conv2d)
           [B, F] -> [B, 1, F] if window_size is None (for Conv1d)
        y: [B] -> [B - H + 1, 1] or [B, NY] -> [B - H + 1, NY] where
        C = 1: channels, B: batch, H: window, F: features, NY: number of outputs

        Use `AsymmetricLosses` for NeuralNet.criterion.

        `estimator.module` may return [B, Ts * NY] or [B, Ts, NY] or [B, NY, Ts].
        If NY = Ts, assumes that [B, NY, Ts] is returned.

        Parameters
        ----------
        estimator : TEstimator
            The estimator to wrap.
        window_size : int | None
            The size of the sliding window.
            Make sure that CNN kernel size is equal or larger than this.
            If None, no sliding window is applied.
        m_type : Literal['mean', 'median', 'nanmean', 'nanmedian'], optional
            M-statistics type to return from `predict` by default, by default "mean"
        var_type : Literal['var', 'std', 'ptp', 'nanvar', 'nanstd'], optional
            Variance type to return from `predict` by default, by default "std"

        Examples
        --------
        >>> from sklearn_utilities.torch.skorch import SkorchCNNReshaperProba, AsymmetricLosses
        >>> from skorch import NeuralNet
        >>> import torch.nn as nn
        >>> net = nn.Sequential(nn.Conv1d(1, 10, 3), nn.GELU(), nn.Flatten(), nn.LazyLinear(5))
        >>> est = SkorchCNNReshaperProba(NeuralNet(module=net, criterion=AsymmetricLosses(ts=5)))
        >>> est.fit(X_train, Y_train)
        >>> y_pred, y_std = est.predict(X_test, return_std=True)
        """
        super().__init__(
            estimator, m_type=m_type, var_type=var_type, window_size=window_size
        )



from __future__ import annotations

from typing import Any, Generic, Literal, overload

from typing_extensions import Self

from ..estimator_wrapper import EstimatorWrapperBase
from ..types import TX, TY, TEstimator, TEstimatorVar
from .dummy_regressor import DummyRegressorVar


class ComposeVarEstimator(
    EstimatorWrapperBase[TEstimator], Generic[TEstimator, TEstimatorVar]
):
    """Compose an estimator with a variance estimator."""

    def __init__(
        self, estimator: TEstimator, estimator_var: TEstimatorVar = DummyRegressorVar()
    ) -> None:
        """Compose an estimator with a variance estimator.

        Parameters
        ----------
        estimator : TEstimator
            The estimator to be wrapped.
        estimator_var : TEstimatorVar, optional
            The variance estimator to be wrapped, by default DummyRegressorVar()
        """
        super().__init__(estimator)
        self.estimator_var = estimator_var

    def fit(self, X: TX, y: TY, **fit_params: Any) -> Self:
        self.estimator.fit(X, y, **fit_params)
        self.estimator_var.fit(X, y, **fit_params)
        return self

    @overload
    def predict(
        self, X: TX, return_std: Literal[False] = ..., **predict_params: Any
    ) -> TY:
        ...

    @overload
    def predict(
        self, X: TX, return_std: Literal[True], **predict_params: Any
    ) -> tuple[TY, TY]:
        ...

    def predict(
        self, X: TX, return_std: bool = False, **predict_params: Any
    ) -> TY | tuple[TY, TY]:
        if return_std:
            return (
                self.estimator.predict(X, **predict_params),
                self.estimator_var.predict(X, return_std=True, **predict_params)[1],
            )
        return self.estimator.predict(X, **predict_params)

    def predict_var(self, X: TX, **predict_params: Any) -> TY:
        return self.estimator_var.predict_var(X, **predict_params)



from __future__ import annotations

from typing import Literal

import numpy as np
from numpy import ndarray
from numpy.typing import ArrayLike
from pandas import DataFrame, Series
from sklearn.dummy import DummyRegressor
from typing_extensions import Self

from ..utils import drop_X_y


class DummyRegressorVar(DummyRegressor):
    """DummyRegressor with 1.0 variance."""

    def __init__(
        self,
        *,
        strategy: Literal["mean", "median", "quantile", "constant", "mean"] = "mean",
        constant: float | None | ArrayLike | int = None,
        quantile: float | None = None,
        allow_nan: bool = True,
    ) -> None:
        super().__init__(strategy=strategy, constant=constant, quantile=quantile)
        self.allow_nan = allow_nan
        self._var_regressor = DummyRegressor(strategy="constant", constant=1.0)

    def fit(
        self,
        X: ArrayLike,
        y: ArrayLike,
        sample_weight: ArrayLike | None = None,
    ) -> Self:
        if (
            self.allow_nan
            and isinstance(y, (Series, DataFrame))
            and isinstance(X, (Series, DataFrame))
        ):
            X, y = drop_X_y(X, y)
        self._var_regressor = DummyRegressor(strategy="constant", constant=np.var(y))
        self._var_regressor.fit(X, y, sample_weight)
        return super().fit(X, y, sample_weight)

    def predict(
        self, X: ArrayLike, return_std: bool = False
    ) -> ndarray | tuple[ndarray, ndarray]:
        return super().predict(X, False), self.predict_var(X)

    def predict_var(self, X: ArrayLike) -> ArrayLike:
        return self._var_regressor.predict(X)
